import { GameObject3D, IGameObject3D } from '../types/game-object';
import {
  BufferGeometry,
  Color,
  ColorRepresentation,
  Mesh,
  MeshPhongMaterial,
} from 'three';
import { assetManager } from '../assets/manager';
import { AssetInfo } from '../types/asset';
import { ExposeProperty } from '../decorators/property';

export interface IPhysicalObject extends IGameObject3D {
  color: ColorRepresentation;
  transparency: number;
  texture?: string;

  /**
   * Get the geometry
   */
  getGeometry(): BufferGeometry;

  /**
   * Get current mesh
   */
  getMesh(): Mesh | undefined;

  /**
   * (re)Initialize the mesh
   */
  initializeMesh(): void;
}

export class PhysicalObject extends GameObject3D implements IPhysicalObject {
  protected geometry?: BufferGeometry;
  protected material?: MeshPhongMaterial;
  protected mesh?: Mesh;
  private _texture?: string;
  private _color = new Color(0xffffff);
  private _transparency = 1;

  @ExposeProperty({ type: Color })
  get color() {
    return this.material?.color || this._color;
  }
  set color(color: ColorRepresentation) {
    this._color = new Color(color);
    if (!this.material) return;
    this.material.color = this._color;
  }

  @ExposeProperty({ type: Number })
  get transparency() {
    return this.material ? 1 - this.material.opacity : this._transparency;
  }
  set transparency(value: number) {
    this._transparency = value;
    if (!this.material) return;
    this.material.transparent = this._transparency != 0;
    this.material.opacity = 1 - this._transparency;
    this.material.needsUpdate = true;
  }

  @ExposeProperty({ type: AssetInfo })
  get texture() {
    return this._texture;
  }
  set texture(path: string | undefined) {
    if (!path) {
      this._texture = undefined;
      if (!this.material) return;
      this.material.map = null;
      this.material.needsUpdate = true;
      return;
    }

    const asset = assetManager.getAssetByPath(path);
    if (!asset || !asset.texture) {
      console.error(`Asset ${path} does not exist or is not loaded`);
      return;
    }

    this._texture = path;
    if (!this.material) return;
    this.material.map = asset.texture;
    this.material.needsUpdate = true;
  }

  constructor() {
    super();
    this.name = this.objectType;
  }

  getGeometry() {
    return this.geometry!;
  }

  getMesh() {
    return this.mesh;
  }

  initializeMesh() {
    if (!this.geometry) return;
    if (this.mesh) {
      this.remove(this.mesh);
    }
    this.material = new MeshPhongMaterial();
    this.mesh = new Mesh(this.geometry, this.material);
    this.add(this.mesh);
  }
}
