import { Color, Euler, Object3D, Vector3 } from 'three';
import { Property } from './property';
import { ExposeProperty } from '../decorators/property';
import { readMetadataOf } from '../utils/read-metadata';

export interface IGameObject {
  objectType: string;
  virtual: boolean;
  uuid: string;
  name: string;
  visible: boolean;
  archivable: boolean;
  properties: Property[];
  serialize(): SerializedObject;
  copy(object: Object3D, recursive?: boolean): this;
  setOwnProperty(key: keyof GameObject, value: unknown): void;
  setOwnProperty(key: string, value: unknown): void;
  setOwnProperty(key: string, value: unknown): void;
  deserialize(input: SerializedObject): void;
}

export interface IGameObject3D extends IGameObject {
  locked: boolean;
  position: Vector3;
  scale: Vector3;
  rotation: Euler;
}

export interface IGameObjectConstructor<T = IGameObject> {
  new (...args: any[]): T;
}

export class GameObject extends Object3D implements IGameObject {
  /**
   * Type of this Game Object.
   */
  public objectType = 'GameObject';

  /**
   * Virtual game objects are usually built-in types that
   * do not have physical representations in the game.
   *
   * These types still extend Object3D so they could
   * have effects in the 3D world.
   */
  public virtual = false;

  @ExposeProperty({ type: String, exposed: false, readonly: true })
  public override uuid!: string;

  @ExposeProperty({ type: String })
  public override name: string = '';

  @ExposeProperty({ type: Boolean })
  public override visible: boolean = true;

  /**
   * Object gets included in exported files if `true`.
   * @default true
   */
  @ExposeProperty({ type: Boolean })
  public archivable: boolean = true;

  constructor() {
    super();
    this.name = this.objectType;
  }

  private get excludedProperties() {
    return readMetadataOf<string>(this, 'excludedProperties');
  }

  /**
   * The exposed properties for this game object, used for the editor
   */
  get properties() {
    const exclude = this.excludedProperties;
    const properties = readMetadataOf<Property>(this, 'properties');
    return properties.filter((item) => !exclude.includes(item.name!));
  }

  /**
   * Serialize GameObject for exporting
   */
  serialize() {
    const object: SerializedObject = {
      name: this.name,
      objectType: this.objectType,
      children: this.children
        .filter(
          (entry) => entry instanceof GameObject && entry.archivable !== false
        )
        .map((entry) => (entry as GameObject).serialize()),
      visible: this.visible,
    };

    const keys = this.properties.map((property) => property.name!);

    Object.assign(
      object,
      keys.reduce<{ [x: string]: unknown }>((obj, key) => {
        const indexable = this as Record<string, unknown>;
        const value = (indexable[key] as any)?.toArray
          ? (indexable[key] as any).toArray()
          : indexable[key];

        return {
          ...obj,
          [key]: value,
        };
      }, {})
    );

    return object;
  }

  override copy(object: Object3D, recursive = true) {
    super.copy(object as any, recursive);
    this.properties
      .map((property) => property.name!)
      .filter((key) => !['position', 'rotation', 'scale', 'uuid'].includes(key))
      .forEach((key) => this.setOwnProperty(key, (object as any)[key]));
    return this;
  }

  /**
   * This function can be used to modify any property of the object by key.
   * @param key Key to set
   * @param value Value to set
   */
  public setOwnProperty(key: keyof GameObject, value: unknown): void;
  public setOwnProperty(key: string, value: unknown): void;
  public setOwnProperty(key: string, value: unknown) {
    const indexable = this as any;
    if (indexable[key]?.fromArray && Array.isArray(value)) {
      indexable[key].fromArray(value);
    } else if (indexable[key]?.isColor) {
      indexable[key] = new Color(value as string);
    } else if (indexable[key]?.copy) {
      indexable[key].copy(value);
    } else {
      indexable[key] = value;
    }
  }

  /**
   * Deserialize a serialized object into properties on this object.
   * @param input Serialized information
   */
  deserialize(input: SerializedObject) {
    Object.keys(input)
      .filter((key) => !['children', 'objectType'].includes(key))
      .forEach((key) => this.setOwnProperty(key, input[key]));
  }
}

export class GameObject3D extends GameObject implements IGameObject3D {
  public objectType = 'GameObject3D';

  @ExposeProperty({ type: Boolean })
  public locked = false;

  @ExposeProperty({ type: Vector3 })
  public override position!: Vector3;

  @ExposeProperty({ type: Vector3 })
  public override scale!: Vector3;

  @ExposeProperty({ type: Euler })
  public override rotation!: Euler;
}

export interface SerializedObject {
  [x: string]: unknown;
  name: string;
  objectType: string;
  children: SerializedObject[];
  visible: boolean;
}
